
;   F F F F F     S S S     E E E E E           A         S S S     MM     MM
;   F           S       S   E                 A   A     S       S   M M   M M
;   F           S           E               A       A   S           M  M M  M
;   F F F F       S S S     E E E E         A       A     S S S     M   M   M
;   F                   S   E               A A A A A           S   M       M
;   F           S       S   E               A       A   S       S   M       M
;   F             S S S     E E E E E   o   A       A     S S S     M       M


; ------------------------------------------------------------------------
;                           What is FSE?
; ------------------------------------------------------------------------

;FSE is a full-screen editor for MS-DOS.  It's similar to GWBASIC's, in that
;it allows the user to cursor-up and execute commands displayed further up 
;the screen.

;FSE is a TSR - that is, it is a resident program that installs itself
;as part of DOS.

;FSE.COM and FSE.ASM (the full source code) are released into the public
;domain by Robert Schifreen.

; ------------------------------------------------------------------------
;                    Development Environment
; ------------------------------------------------------------------------
 
;                Machine:  Epson PC
;       Operating System:  PC-DOS 3.30.  Will run under DOS 2.x or 3.x 
;               Language:  8088 assembly language
;              Assembler:  Microsoft MACRO Assembler version 5.00
;                 Linker:  Microsoft 8086 Object Linker version 3.06
;         Program Format:  Must be run as a .COM file, not .EXE
; Editor for source file:  Word Perfect Program Editor version 4.2

; ------------------------------------------------------------------------
;                         Use of interrupts
; ------------------------------------------------------------------------

; The program takes over the handling of Int 16h for functions 0,1,2.  
; If a function > 2 is requested, the program jumps to the old code.
; For functions <=2, the old code is NOT called at all.


;Interrupt 21h is also taken over.  The logic used is:

;  clear flag in case it was set
;  If ah = 0Ah to request get-string function then:
;           Set flag to show that we are executing a func 0Ah.
;           jump to old int 21h code
;   else
;           jump to old int 21h code but don't set flag

;The Int 16h handler checks the flag, and only allows the user to enter the
;full-screen editor if it is set.

; ------------------------------------------------------------------------
;                        Using the Full Screen Editor
; ------------------------------------------------------------------------

;The key to enter the editor from DOS is up-arrow.

;FSE is only active when you are at the DOS prompt. In applications progs,
;pressing up-arrow will not enter the FSE editor (unless that program uses
;the Int 21h Func 0Ah get-string function, like COMMAND.COM does.

;See the help text in the install procedure for list of keys.

;Any alphabetic or numeric key adds that char to screen, in a way that
;reflects current insert state.  Insert state for FSE can only be toggled
;while you are in the editor.

; ------------------------------------------------------------------------
;                     FSE and other resident programs
; ------------------------------------------------------------------------

; FSE has been tested with a number of programs, and works happily with many,
; including SideKick.  It will not work with Dosedit.

; This program will happily co-exist with ALIAS if FSE is loaded last.

; ------------------------------------------------------------------------
;                           Use of DOS and BIOS
; ------------------------------------------------------------------------

; FSE is not hardware dependent.  All cursor movement/reading is done through
; BIOS calls to interrupt 10h.  Also, bp is preserved during a call to
; this interrupt, as a bug in some IBM BIOSes can corrupt this register.

; DOS calls are used where possible to display text.  This includes the
; help screen displayed when FSE is typed.  So, the help screen can be
; redirected to a file.

; HOWEVER, it is not possible to use a DOS call to display ascii character
; 1Bh (a left pointing arrow) as ANSI.SYS will intercept it as part of a
; possible ESCape sequence.  Because of this, a BIOS call is used to display
; this character, and it will appear on the screen (not on the standard output
; device) when the help screen is displayed.

; Note that the right pointing arrow has a code of Ctrl-Z, and will be seen
; as an end-of-file marker by DOS, so you can't get a printout of FSE's
; help screen by redirecting FSE to a file and printing it.  Instead, run
; FSE on the screen and then execute PrtSc to copy to the printer.

; ------------------------------------------------------------------------
;                       Installing FSE
; ------------------------------------------------------------------------

;Note that FSE.COM should not be renamed.  If you do rename the program,
;ensure that the new name is 3 characters long (plus the .COM extension)
;or the /N and /F flags won't work.

;Type FSE to install the program.  A help screen will be displayed also.  
;If FSE is already installed, the program won't normally allow you to load
;another copy and a message to that effect will be displayed.

;However, if you want to force another copy to be loaded (eg, if you have 
;loaded a resident program that conflicts with FSE and you want FSE to take
;over its interrupts again, this can be achieved by typing FSE /F (note the
;space after FSE).

;Full instructions are displayed when FSE is loaded.  If you don't want
;the instructions displayed (because you want to cursor over a long line
;that is currently on screen but will be overlaid with the instructions),
;type FSE /N for no instructions.  If you do this, typing FSE at a later
;date without the /N will show the instructions.
;Note that you can't use both switches at once - doing a /N does an implied
;/F as well, though.


; ------------------------------------------------------------------------
;                       History Of This Version
; ------------------------------------------------------------------------

;  1.51    Int 16h calls > 2 are passed to old int 16h vector
;  1.50    Added /N switch
;  1.40    Previous versions only activated FSE if an Int 21h func 0Ah call
;          was in progress, AND the user was at the dos prompt.  Now, just 
;          being in the midst of a func 0Ah call is enough.  This means that
;          FSE will work within all progs that use this way of reading the
;          keyboard, eg Infocom games and DEBUG.
;  1.31    Ctrl-W deletes up to space or comma, and not just space.
;  1.30    Initial version 13 June 1987

; ------------------------------------------------------------------------
;                          Macro Definitions
; ------------------------------------------------------------------------

push_m  macro arglist
              irp      register,<arglist>
              push     register
              endm
        endm
;
pop_m   macro arglist
              irp      register,<arglist>
              pop      register
              endm
        endm

show    macro string
               push  ax
               push  dx
               mov   ah,9
               mov   dx,offset cs:string
               int   21h
               pop   dx
               pop   ax
        endm

video_int   macro                ; preserve bp during an Int 10h call, as a
               push  bp          ; bug in the IBM BIOS can corrupt this reg.
               int   10h
               pop   bp
            endm

disp_left   macro                ; display left arrow on screen.  This has to
                                 ; be done through BIOS, cos ANSI.SYS driver
                                 ; will interpret the character (1Bh) as ESC
                                 ; if DOS calls are used.


               push   ax
               push   bx
               mov    ah,0eh
               mov    al,LEFT
               mov    bl,7
               mov    bh,0
               video_int
               pop    bx
               pop    ax
          endm
;
; ------------------------------------------------------------------------
;                             Equates
; ------------------------------------------------------------------------
;
CR      equ   0dh                       ; carriage return character
LF      equ   0ah                       ; line feed character
EOM     equ   36                        ; dollar sign is end of message marker
UP      equ   24                        ; up_arrow character
DOWN    equ   25                        ; down arrow
LEFT    equ   27                        ; left arrow
RIGHT   equ   26                        ; right arrow
COMMA   equ   44                        ; comma character
PSP_OFF equ   81h                       ; start of FSE's buffer within old PSP

VER1    equ   '1'                       ; Version number   - units
VER2    equ   '5'                       ;                  - tenths
VER3    equ   '1'                       ;                  - hundredths

;
; ------------------------------------------------------------------------
;                          Start of The Program
; ------------------------------------------------------------------------
;

code    segment
        assume  cs:code,ds:code,es:nothing
        org     100h

start:  jmp     init                    ; init code at end of res section

int21_entry:                            ; int 21h comes here.  If it's a
                                        ; call to function 0Ah (get string),
                                        ; we call the func but 1st set flag
                                        ; if at the DOS prompt, so that FSE
                                        ; will be active during operation.


        pushf
        mov     byte ptr cs:prompt_flag,0  ; clear activate-fse flag, in case
                                           ; it was set by last Int 21h call.

        cmp     ah,0ah                  ; is this a get string call?
        je      func_0a_found

orig_int21:

        popf

orig_int21_2:


        db      0eah                    ; not a func 0ah, so continue with 
                                        ; int 21h as normal. EAh is far jump.
int21_ip        dw 0
int21_cs        dw 0

func_0a_found:

        popf                            ; left over from before

                                        ; save current cursor position for use
        push_m <ax,bx,cx,dx>            ; by TAB_KEY proc.
        mov    ah,3
        mov    bh,0
        video_int                       ; get column --> dl
        mov    byte ptr cs:after_prompt,dl ; save position
        pop_m  <dx,cx,bx,ax>

                                        ; function 0Ah is being called, so...

                                        ; 1. set the flag to activate FSE
                                        ; 2. resume Int 21h call.

                                        ; Flag will be cleared at top of
                                        ; next Int 21h call.

        mov     byte ptr cs:prompt_flag,1 ; activate FSE
        jmp     orig_int21_2


;
; ------------------------------------------------------------------------
;                         End Of Int 21h Handler
; ------------------------------------------------------------------------
;


;
; ------------------------------------------------------------------------
;                        Start Of Int 16h Handler
; ------------------------------------------------------------------------
;


intent:                                 ; Interrupt 16h now comes here


        sti	                                   
        cmp     ah,2
        jbe     less_than_3
        jmp     at_func                 ; ah > 2, so jmp to old code

less_than_3:

        push    ds                                 
        push    bx                                 
        mov     bx,40h                  ; set segment
        mov     ds,bx                              
        or      ah,ah                              
        jz      read_key                ; function 0
        dec     ah                                 
        jne     labl
        jmp     is_key_pending          ; function 1

labl:

        dec     ah                                 
        jne     tidy_up
        jmp     get_shift_status        ; function 2

tidy_up:

        pop     bx                                 
        pop     ds                                 
        iret	                                   

read_key:                               ; get char from buffer



        cli
        push    bp
        push    si
        push    ax
        mov     al,cs:pointer           ; Is there a key waiting to be
        mov     ah,0                    ; stuffed into fake 'buffer'?
        mov     bp,ax
        pop     ax
        mov     si,PSP_OFF
        cmp     byte ptr cs:[bp+si],0   ; anything in the stuff buffer
        je      dun_stuffin             ; If not, continue with real func 0.
        mov     al,byte ptr cs:[bp+si]  ; put faked ascii code in AL.  
        pop     si
        pop     bp

        push    dx
        push    bx
        mov     dl,al                   ; save it
        mov     bx,offset cs:scans
        xlat                            ; get scan code for ascii code
        mov     ah,al                   ; put scan code in ah
        mov     al,dl                   ; and put ascii code back in al
        pop     bx
        pop     dx
        inc     cs:pointer              ; advance ptr to next stuffable char
        jmp     short tidy_up

dun_stuffin:                            ; Nothing to stuff so cont as normal

        pop     si
        pop     bp
        mov     bx,word ptr ds:[001ah]  ; pointer to head of BIOS kbd buffer                       
        cmp     bx,word ptr ds:[001ch]  ; if head = tail then nothing there                       
        jnz     entry_in_buffer
        sti
        jmp     read_key                ; loop till key in buffer                            

entry_in_buffer:

        mov     ax,[bx]                 ; get scan code and ascii code                          

                                        ; If we aren't at the DOS prompt,
                                        ; don't check for up-arrow.

        cmp     byte ptr cs:prompt_flag,1
        jne     carry_on2               ; don't check for up-arrow unless
                                        ; flag is set to say we are at the
                                        ; DOS prompt.

        cmp     ax,4800h                ; up arrow pressed, to enter editor?
        jne     carry_on2               ; jmp if not
        call    save_cursor_pos         ; guess what!
        call    editor                  ; This is the full-screen editor
        call    restore_cursor_pos      ; and clean up behind us
        jmp     read_key                ; and back where we belong

carry_on:

        pop     bp
        pop     ax

carry_on2:

        inc     bx                                 
        inc     bx                                 
        mov     word ptr ds:[1ah],bx    ; inc BIOS buffer pointers                         
        cmp     bx,word ptr ds:[0082h]  ; end of buffer?
        je      dont_tidy
        jmp     tidy_up

dont_tidy:

        mov     bx,word ptr ds:[0080h]  ; else loop round again in buffer                         
        mov     word ptr ds:[1ah],bx                          
        jmp     tidy_up                 ; and iret

is_key_pending:

        cli

                                        ; If there's a key pending in the
                                        ; stuff buffer, don't read the real
                                        ; BIOS buffer but stuff from 
                                        ; the buffer instead.

        push    bp
        push    si
        push    ax
        mov     al,cs:pointer
        mov     ah,0
        mov     bp,ax
        pop     ax
        mov     si,PSP_OFF
        cmp     byte ptr cs:[bp+si],0   ; is character at pointer a 0
        je      nowt_pending            ; jump if nothing to stuff

        push    ax                      ; clear Z flag to indicate key pending
        mov     ah,al
        inc     al
        cmp     ah,al
        pop     ax

        mov     al,byte ptr cs:[bp+si]  ; put char from buffer into AL
        pop     si
        pop     bp
        push    dx
        push    bx
        mov     dl,al                   ; save it
        mov     bx,offset cs:scans
        xlat                            ; get scan code
        mov     ah,al                   ; put scan code in ah
        mov     al,dl                   ; and put ascii code back in al
        pop     bx
        pop     dx
        jmp     short stuffed           ; And we're all done here.

nowt_pending:

        pop     si
        pop     bp
        mov     bx,word ptr ds:[001ah]  ; is head of BIOS buffer = tail?
        cmp     bx,word ptr ds:[001ch]  ; set Z flag as sign to caller
        mov     ax,[bx]                            

stuffed:

        sti	                                   
        pop     bx                                 
        pop     ds                                 
ret_f:  db 0cah,2,0  ; retf 2           ; Must be RETF.  Can't leave assembler
                                        ; to do this, or it will become RET 2.

get_shift_status:
                                        ; Do Int 16h function 2
        mov     al,byte ptr ds:[0017h]  ; get byte from low mem

                                        ; As a sign to the install routine,
                                        ; int 16h function 2 increments the
                                        ; first byte of the ICA, at 0:4F0h

        inc     byte ptr ds:0f0h        ; ds is already at 40h

        jmp     tidy_up 


at_func:    db  0EAh  ; far jump
old_off     dw  0     ; offset of old int 16 vector
old_seg     dw  0     ; segment of old int 16h vector
       

; ------------------------------------------------------------------------
;                            Procedures
; ------------------------------------------------------------------------

editor proc                             ; This is the main loop of the full
                                        ; screen editor, entered when the user
                                        ; presses up-arrow.  This routine does
                                        ; little more than look for valid key
                                        ; presses and call procedures to deal
                                        ; with those keys.

top_of_editor:

        cmp     byte ptr cs:crsr_size_changed,1
        je      crsr_size_ok
        cmp     byte ptr cs:insert_state,0
        je      crsr_size_ok            ; no INS so don't alter crsr size

        push_m <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; crsr start/end into ch,cl. 
        mov     word ptr cs:crsr_size,cx; save it
        mov     byte ptr cs:crsr_size_changed,1 
        dec     ch
        dec     ch                      ; make cursor 2 scan lines higher
        mov     ah,1
        video_int
        pop_m   <dx,cx,bx,ax>

crsr_size_ok:

        push    ax
        call    get_key                 ; get scan code,ascii code into ah,al.
                                        ; Don't alter real BIOS buffer in
                                        ; segment 40h.

        cmp     ax,4800h                ; Cursor-up?
        jne     try2
        call    cursor_up
        pop     ax
        jmp     top_of_editor

try2:

        cmp     ax,4b00h                ; crsr left
        jne     try3
        call    cursor_left
        pop     ax
        jmp     top_of_editor

try3:

        cmp     ax,4d00h                ; crsr right
        jne     try4
        call    cursor_right
        pop     ax
        jmp     top_of_editor

try4:

        cmp     ax,5000h                ; crsr down
        jne     try5
        call    cursor_down
        pop     ax
        jmp     top_of_editor

try5:



        cmp     ax,5300h                ; DEL key?
        jne     try6
        call    cursor_off
        call    delete_character
        call    cursor_on
        pop     ax
        jmp     top_of_editor

try6:

        cmp     ax,0E08h                ; Backspace key?
        jne     try7
        call    cursor_off
        call    cursor_left             ; backsp = crsr left then DEL 
        call    delete_character
        call    cursor_on
        pop     ax
        jmp     top_of_editor

try7:

        cmp     ax,011Bh                ; ESC pressed to abandon editor?
        jne     try8
        pop     ax
        jmp     finish

try8:

        cmp     ax,5200h                ; INS key pressed
        jne     try9
        call    toggle_insert_state
        pop     ax
        jmp     top_of_editor

try9:
        cmp     ax,1205h                ; Ctrl-E pressed?
        jne     try10
        call    erase_eol               ; erase to end of line
        pop     ax
        jmp     top_of_editor

try10:  
        cmp     ax,4700h                ; HOME key pressed?
        jne     try11
        call    cursor_off
        call    home_crsr               ; go to start of line
        call    cursor_on
        pop     ax
        jmp     top_of_editor

try11:
        cmp     ax,4F00h                ; END key pressed?
        jne     try12
        call    cursor_off
        call    end_crsr                ; go to end of line
        call    cursor_on
        pop     ax
        jmp     top_of_editor
 
try12:

        cmp     ax,2106h                ; Ctrl-F
        jne     try13
        call    word_forward            ; move crsr forward one word
        pop     ax
        jmp     top_of_editor


try13:

        cmp     ax,3002h                ; Ctrl-B
        jne     try14
        call    word_backward           ; move crsr back one word
        pop     ax
        jmp     top_of_editor


try14:

        cmp     ax,1117h                ; Ctrl-W
        jne     try15
        call    cursor_off
        call    del_word                ; delete current word
        call    cursor_on
        pop     ax
        jmp     top_of_editor


try15:

        cmp    ax,4900h                 ; PgUp
        jne    try16
        call   page_up
        pop    ax
        jmp    top_of_editor

try16:

        cmp    ax,5100h                 ; PgDn
        jne    try17
        call   page_down
        pop    ax
        jmp    top_of_editor

try17:

        cmp    ax,0f09h                 ; TAB key
        jne    try18
        call   tab_key
        pop    ax
        jmp    top_of_editor

try18:

        cmp    ax,1c0dh                 ; Carriage return?
        je     end_of_editor
      
look_for_real_char:

        cmp     al,32                   ; is char a valid printable char?
        jb      no_more_tries 
        cmp     al,126
        ja      no_more_tries
                                        ; char is a printable 
                                        ; char so display it..

        cmp     byte ptr cs:insert_state,0   ; is insert on?
        je      no_ins_needed                ; no.
        call    cursor_off
        call    insert                  ; shift line right by 1 char
        call    cursor_on
        mov     byte ptr cs:flag,1      ; say that no crsr right needed

no_ins_needed:

        mov     ah,0ah                  ; display char
        mov     bh,0                    ; page
        mov     bl,7                    ; attribute. not really needed
        mov     cx,1                    ; no of times to display
        video_int                       ; write the char
        call    cursor_right            ; this is not automatic with func 0Ah
        pop     ax                      ; tidy stack
        jmp     top_of_editor

no_more_tries:

        pop     ax
        jmp     top_of_editor           ; If none of the above, try again.
 
end_of_editor:                          ; CR was pressed to exit the editor

        pop     ax                      ; tidy up the stack
        mov     byte ptr cs:prompt_flag,0  ; can turn off editor now
                                        ; Take the entire command line that the
                                        ; cursor is currently on, and copy it
                                        ; to the PSP buffer ready for
                                        ; faking into the keyboard buffer.

        push_m  <ax,bx,cx,dx,bp,si>     ; save the registers we'll use
        mov     bp,PSP_OFF              ; start of buffer for chars
        mov     si,0                    ; pointer into the buffer

        mov     ah,3                    ; read current crsr pos
        mov     bh,0
        video_int
                                        ; DH/DL now contains crsr position
                                        ; from where we start scanning the
                                        ; screen.  Currently, this position
                                        ; is wherever the crsr happens to be.
                                        ; By putting 0 in dl, scanning would
                                        ; start from the far left of the line.
                                        ; Whatever happens, scanning stops at
                                        ; end of screen line (80 cols).

again:

        mov     ah,8                    ; read the char that's there
        mov     bh,0
        video_int
        mov     cs:[bp+si],al           ; put char into buffer
        inc     si                      ; inc buffer pointer
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get current crsr position
        cmp     dl,79                   ; End of line?
        pop_m   <dx,cx,bx,ax>
        je      done                    ; jump if yes
        call    cursor_right            ; else move the cursor right
        jmp     short again             ; and start again

done:

        pop_m   <si,bp,dx,cx,bx,ax>     ; restore registers


                                        ; The characters to be stuffed into 
                                        ; the keyboard are now in the buffer.
                                        ; Put the pointer to start of buffer.

        mov     cs:pointer,0            ; point to start of scanned string

                                        ; The scanned string is in the buffer.
                                        ; Before we leave, we must strip off 
                                        ; the trailing spaces and make them 
                                        ; into zeros. Also, terminate the
                                        ; remaining string with a CR.

        push_m  <bp,si,ax>
        mov     bp,80
        mov     si,PSP_OFF

freda:

        mov     ah,byte ptr cs:[bp+si]  ; scan buffer, starting from end
        cmp     ah,' '                  ; We're done if it's neither space..
        je      keep_on
        cmp     ah,0                    ; .. nor null.
        je      keep_on
        inc     bp                      ; adjust pointer
        mov     byte ptr cs:[bp+si],CR  ; and plug CR at end of string
        pop_m   <ax,si,bp>
        jmp     short finish

keep_on:

        mov     byte ptr cs:[bp+si],0   ; Move 1 space back and keep going
        dec     bp
        jmp     freda

finish:
        
        cmp     byte ptr cs:crsr_size_changed,0
        je      no_need_for_this
        push    cx
        mov     cx,word ptr cs:crsr_size
        mov     ah,1
        video_int                       ; tidy up cursor size
        pop     cx
        mov     byte ptr cs:crsr_size_changed,0

no_need_for_this:

        ret                             ; Exit the editor.  The string taken
                                        ; from the screen is ready to be given
                                        ; char-by-char to subsequent Int 16h's.

editor endp
;
delete_character proc

                                        ; Shift everything after the crsr to
                                        ; the left by 1 position.

        push_m  <ax,bx,cx,dx,bp,si>    

        mov     bp,PSP_OFF              ; start of buffer for chars
        mov     si,0                    ; pointer into the buffer

        mov     ah,3                    ; read current crsr pos
        mov     bh,0
        video_int
                                        ; DH/DL now contains crsr position
                                        ; from where we start scanning the
                                        ; screen.  

        mov word ptr cs:ctrl_w_save,dx  ; save cursor position (dx)

again4:

        mov     ah,8                    ; read the char that's there
        mov     bh,0
        video_int
        mov     cs:[bp+si],al           ; put char into buffer
        inc     si                      ; inc buffer pointer
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get current crsr position
        cmp     dl,79                   ; End of line?
        pop_m   <dx,cx,bx,ax>
        je      done4                   ; jump if yes
        call    cursor_right            ; else move the cursor right
        jmp     short again4            ; and start again

done4:

        mov     byte ptr cs:[bp+si],0   ; terminate string in buffer
                                        ; with a null.

        pop_m   <si,bp,dx,cx,bx,ax>


                                        ; the line is now in the buffer,
                                        ; starting from the pos after
                                        ; the cursor.

        call    get_back_cursor
        call    erase_eol               ; delete all chars after crsr
        call    get_back_cursor

        push_m  <ax,bx,cx,dx,bp,si>
        mov     si,PSP_OFF
        mov     bp,1                    ; skip 1st char


                                        ; Copy the buffer from bp onto the 
                                        ; screen, starting at the cursor 
                                        ; position.  Stop copying when null
                                        ; found in buffer.

        mov    bh,0                     ; video page. just specify it once.

char_erased:

        mov     al,byte ptr cs:[bp+si]  ; get char from buffer
        cmp     al,0                    ; end of buffer reached?
        je      char_deleted

        mov     ah,0eh
        video_int                       ; put char on screen and move crsr 
        inc     bp
        jmp     short char_erased

char_deleted:

        mov     byte ptr cs:[si],0      ; put a null at start of buffer to
                                        ; stop Int 16h handler thinking that
                                        ; the junk left in the buffer is for
                                        ; stuffing into keyboard.

        pop_m   <si,bp,dx,cx,bx,ax>
        call    get_back_cursor
        ret
delete_character endp
;
cursor_up       proc                    ; move the cursor up a line. If
                                        ; unable to move, set cs:unable = 1.

        push_m  <ax,bx,cx,dx>
        mov     ah,3                    ; read crsr row,column into dh,dl
        mov     bh,0                    ; page 0
        video_int
        cmp     dh,0       
        jne     continue_on             ; Don't let crsr go above top line

        mov     byte ptr cs:unable,1
        jmp     short cant1

continue_on:

        dec     dh                      ; move cursor up by decrementing line    
        mov     ah,2                    ; set crsr to position dh,dl
        video_int
cant1:

        pop_m   <dx,cx,bx,ax>
        ret

cursor_up endp
;
cursor_down     proc                    ; move the cursor down a line. If crsr
                                        ; unable to move, set cs:unable = 1
        push_m  <ax,bx,cx,dx>
        mov     ah,3                    ; read crsr row,column into dh,dl
        mov     bh,0                    ; page 0
        video_int
        cmp     dh,24
        jne     continue2

        mov     byte ptr cs:unable,1
        jmp     short cant2

continue2:

        inc     dh                      ; move cursor down
        mov     ah,2                    ; set crsr to position dh,dl
        video_int

cant2:

        pop_m   <dx,cx,bx,ax>
        ret
cursor_down     endp
;
cursor_left     proc                    ; Move cursor left 1 char. 
                                        ; If unable to move because cursor
                                        ; is already at left of screen, set
                                        ; cs:unable to 1 before aborting.

        push_m  <ax,bx,cx,dx>
        mov     ah,3                    ; read crsr row,column into dh,dl
        mov     bh,0                    ; page 0
        video_int
        cmp     dl,0                    ; are we already at LHS of screen?
        jne     can_continue            ; no, so go on
        mov     byte ptr cs:unable,1    ; yes, so set flag and abort
        jmp     short cant3

can_continue:

        dec     dl                      ; move cursor left
        mov     ah,2                    ; set crsr to position dh,dl
        video_int

cant3:

        pop_m   <dx,cx,bx,ax>
        ret
cursor_left     endp
;
cursor_right    proc                    ; Move cursor right.If unable to move,
                                        ; set cs:unable = 1.
        push_m  <ax,bx,cx,dx>
        mov     ah,3                    ; read crsr row,column into dh,dl
        mov     bh,0                    ; page 0
        video_int
        cmp     dl,79
        jne     continue4

        mov     byte ptr cs:unable,1
        jmp     short cant4

continue4:

        inc     dl
        mov     ah,2                    ; set crsr to position dh,dl
        video_int

cant4:

        pop_m   <dx,cx,bx,ax>
        ret
cursor_right    endp
;
word_forward    proc                    ; put crsr on 1st char of next word

push_m <ax,bx,cx,dx>
        call    save_crsr_word          ; save current cursor position
        call    cursor_left             ; otherwise,if you start on a space,
                                        ; this routine jumps a word
find_space_still:

        mov     ah,3                    ; read crsr position
        mov     bh,0  
        video_int                       ; get crsr position into dl (horiz)
        cmp     dl,79                   ; if at RHS of screen, abort
        je      done_forward

        call    cursor_right
        mov     ah,8
        mov     bh,0


        video_int                       ; get char at crsr position
        cmp     al,20h                  ; is char a space
        jne     find_space_still

                                        ; char is a space
not_real:

        mov     ah,3
        mov     bh,0
        video_int
        cmp     dl,79
        je      done_forward
        call    cursor_right
        mov     ah,8                    ; get char under crsr
        mov     bh,0
        video_int
        cmp     al,32                   ; keep going till it's not a space
        je      not_real
        jmp     moved_forward

done_forward:

        call    restore_crsr_word

moved_forward:

        pop_m   <dx,cx,bx,ax>
        ret
word_forward    endp
;
word_backward   proc

        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get cursor position.  column --> dl
        cmp     dl,0                    ; has cursor reached start of line?
        je      moved_word_back         ; yes, so abort

        mov     byte ptr cs:unable,0    ; clear flag used by crsr_move procs
        call    cursor_left
        cmp     byte ptr cs:unable,1    ; if call failed then abort
        je      moved_word_back

start_proc:

        mov     byte ptr cs:unable,0    ; clear flag used by crsr_move procs
        call    cursor_left
        cmp     byte ptr cs:unable,1    ; if call failed then abort
        je      moved_word_back
        mov     ah,8
        mov     bh,0
        video_int                       ; get char under cursor
        cmp     al,20h                  ; is cursor on a space?
        jne     carry_on_back           ; no, then carry on
        call    cursor_right
        jmp     moved_word_back         ; done

carry_on_back:

        cmp     byte ptr cs:unable,1
        je      moved_word_back
        jmp     short start_proc

moved_word_back:

        pop_m   <dx,cx,bx,ax>
        ret
word_backward   endp
;
del_word proc                           ; Delete current word

        push_m  <ax,bx,cx,dx,bp,si>    

        mov     bp,PSP_OFF              ; start of buffer for chars
        mov     si,0                    ; pointer into the buffer

        mov     ah,3                    ; read current crsr pos
        mov     bh,0
        video_int
                                        ; DH/DL now contains crsr position
                                        ; from where we start scanning the
                                        ; screen.  

        mov word ptr cs:ctrl_w_save,dx  ; save cursor position (dx)

again2:

        mov     ah,8                    ; read the char that's there
        mov     bh,0
        video_int
        mov     cs:[bp+si],al           ; put char into buffer
        inc     si                      ; inc buffer pointer
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get current crsr position
        cmp     dl,79                   ; End of line?
        pop_m   <dx,cx,bx,ax>
        je      done2                   ; jump if yes
        call    cursor_right            ; else move the cursor right
        jmp     short again2            ; and start again

done2:

        mov     byte ptr cs:[bp+si],0   ; terminate string in buffer
                                        ; with a null.

        pop_m   <si,bp,dx,cx,bx,ax>


                                        ; the line is now in the buffer,
                                        ; starting from the pos after
                                        ; the cursor.

        call    get_back_cursor
        call    erase_eol               ; delete all chars after crsr
        call    get_back_cursor

        push_m  <ax,bx,cx,dx,bp,si>
        mov     si,PSP_OFF
        mov     bp,0                    ; point to start of buffer

keep_going:

        mov     ah,byte ptr cs:[bp+si]  ; get character from buffer
        cmp     ah,20h                  ; is it a space?
        je      it_is                   ; jump if it is
        cmp     ah,COMMA                ; same goes for a comma
        je      it_is
        inc     bp
        jmp     short keep_going

it_is:

                                        ; Now skip any more spaces or commas

        inc     bp
        mov     ah,byte ptr cs:[bp+si]  ; get char from buffer
        cmp     ah,20h                  ; is it a space?
        jne     it_is2                  ; no
        jmp     short it_is             ; yes - keep scanning buffer
       
it_is2:

        cmp     ah,COMMA
        jne     it_is3
        jmp     short it_is

it_is3:
                                        ; Copy the buffer from bp onto the 
                                        ; screen, starting at the cursor 
                                        ; position.  Stop copying when null
                                        ; found in buffer.

        mov    bh,0                     ; video page. just specify it once.

erase1:

        mov     al,byte ptr cs:[bp+si]  ; get char from buffer
        cmp     al,0                    ; end of buffer reached?
        je      word_erased

        mov     ah,0eh
        video_int                       ; put char on screen and move crsr 
        inc     bp
        jmp     short erase1

word_erased:

        mov     byte ptr cs:[si],0      ; put a null at start of buffer to
                                        ; stop Int 16h handler thinking that
                                        ; the junk left in the buffer is for
                                        ; stuffing into keyboard.

        pop_m   <si,bp,dx,cx,bx,ax>
        call    get_back_cursor
        ret
del_word endp
;
get_back_cursor proc                    ; Cursor's position is saved when
                                        ; Ctrl-W pressed.  This proc puts crsr
                                        ; back where it was when Ctrl-W was
                                        ; pressed.

        push_m <ax,bx,cx,dx>
        mov    ah,2
        mov    bh,0
        mov    dx,word ptr cs:ctrl_w_save
        video_int
        pop_m  <dx,cx,bx,ax>
        ret
get_back_cursor endp
;
page_up         proc                    ; put cursor on top line of screen
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get cursor row/col into dh/dl
        mov     dh,0                    ; set row = 0, for top of screen
        mov     ah,2
        video_int                       ; set crsr position to new value
        pop_m   <dx,cx,bx,ax>
        ret
page_up         endp
;
page_down       proc                    ; put cursor on bottom line of screen
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get cursor row/col into dh/dl
        mov     dh,24                   ; set row = 25, for bottom of screen
        mov     ah,2
        video_int                       ; set crsr position to new value
        pop_m   <dx,cx,bx,ax>
        ret
page_down       endp
;
tab_key         proc                    ; When the user presses TAB, the crsr
                                        ; is placed over the first character
                                        ; after the prompt.  FSE does this by
                                        ; hooking into int 21h func 0ah, which
                                        ; gets the user's DOS command from the
                                        ; keyboard, and noting the cursor's
                                        ; position when this func is called. 
                                        ; This position will be the first char
                                        ; after the prompt.  If the user 
                                        ; changes the prompt, the new position
                                        ; will be used from then on.  The side
                                        ; effect is that prompts further up 
                                        ; the screen will be assumed to be the
                                        ; same length as the current one, which
                                        ; may not be true.  The code to remedy
                                        ; this is not worth the bother, as most
                                        ; people choose a prompt and stick 
                                        ; to it.

        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; crsr pos into dh/dl
        mov     dl,cs:after_prompt      ; correct dl
        mov     ah,2
        video_int                       ; and set crsr to new position
        pop_m   <dx,cx,bx,ax>
        ret
tab_key         endp
;
insert          proc                    ; shift line right by 1 char, starting
                                        ; at char after crsr. Then put crsr 
                                        ; back where it was.

        push_m  <ax,bx,cx,dx,bp,si>    

        mov     ah,3                    ; read current crsr pos
        mov     bh,0
        video_int
                                        ; DH/DL now contains crsr position
                                        ; from where we start scanning the
                                        ; screen.  

        mov     bp,PSP_OFF              ; start of buffer for chars
        mov     si,0                    ; pointer into the buffer

        mov word ptr cs:ctrl_w_save,dx  ; save cursor position (dx)

again3:

        mov     ah,8                    ; read the char that's there
        mov     bh,0
        video_int
        mov     cs:[bp+si],al           ; put char into buffer
        inc     si                      ; inc buffer pointer
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get current crsr position
        cmp     dl,79                   ; End of line?
        pop_m   <dx,cx,bx,ax>
        je      done3                   ; jump if yes
        call    cursor_right            ; else move the cursor right
        jmp     short again3            ; and start again

done3:

        dec     si
        mov     byte ptr cs:[bp+si],0   ; terminate string in buffer
                                        ; with a null.

        pop_m   <si,bp,dx,cx,bx,ax>


                                        ; the line is now in the buffer,
                                        ; starting from the pos after
                                        ; the cursor.

        call    get_back_cursor
        call    erase_eol               ; delete all chars after crsr

        push_m <ax,bx,cx,dx>
        mov    ah,2
        mov    bh,0
        mov    dx,word ptr cs:ctrl_w_save  ; get old crsr position
        inc    dl                       ; move crsr right
        video_int
        pop_m <dx,cx,bx,ax>
 

                                        ; now copy chars back from buffer.

        push_m  <ax,bx,cx,dx,bp,si>
        mov     si,PSP_OFF
        mov     bp,0                    ; point to start of buffer
        mov     bh,0                    ; video page. just specify it once.

ins1:

        mov     al,byte ptr cs:[bp+si]  ; get char from buffer
        cmp     al,0                    ; end of buffer reached?
        je      insert_done
        mov     ah,0eh
        video_int                       ; put char on screen and move crsr 
        inc     bp
        jmp     short ins1

insert_done:

        mov     byte ptr cs:[si],0      ; put a null at start of buffer to
                                        ; stop Int 16h handler thinking that
                                        ; the junk left in the buffer is for
                                        ; stuffing into keyboard.

        pop_m   <si,bp,dx,cx,bx,ax>
        call    get_back_cursor
        ret
insert          endp
;
erase_eol       proc                    ; Erase to end of current line
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get crsr position
        mov     ah,0Ah                  ; write spaces to clear the line
        mov     al,20h                  ; space
        mov     cx,80
        mov     dh,0
        sub     cx,dx                   ; no of spaces needed = 80-column_pos
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
erase_eol       endp
;
home_crsr       proc                    ; put cursor to start of line
        push_m  <ax,bx,cx,dx>
        mov     ah,3                    ; get crsr position
        mov     bh,0
        video_int
        mov     dl,0                    ; set column = 0
        mov     ah,2
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
home_crsr       endp
;
end_crsr        proc
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get crsr position
        mov     ah,2
        mov     dl,79                   ; put crsr at end of line
        video_int

find_end:

        mov     ah,8                    ; get char under crsr
        video_int
        cmp     al,20h                  ; is it a space?
        jne     found_end               ; if not, we are at end of the line
        call    cursor_left             ; else move cursor and go round again
        jmp     short find_end

found_end:
        call    cursor_right            ; put crsr after last char, not on it
        pop_m   <dx,cx,bx,ax>
        ret
end_crsr        endp
;
save_cursor_pos proc                    ; Save crsr pos before entering editor
        push_m  <dx,cx,bx,ax>
        mov     ah,3                    ; get crsr pos
        mov     bh,0                    ; page 0
        video_int
        mov     word ptr cs:cursor_pos,dx ; save position
        pop_m   <ax,bx,cx,dx>
        ret 
save_cursor_pos endp
;
restore_cursor_pos proc
        push_m  <ax,bx,cx,dx>
        mov     ah,2                    ; set crsr pos
        mov     bh,0                    ; page 0
        mov     dx,cs:cursor_pos        ; get saved position
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
restore_cursor_pos endp
;
save_cursor_before_del proc             ; Save crsr pos before processing DEL

        push_m  <dx,cx,bx,ax>
        mov     ah,3                    ; get crsr pos
        mov     bh,0                    ; page 0
        video_int
        mov     word ptr cs:crsr_pos_del,dx ; save position
        pop_m   <ax,bx,cx,dx>
        ret 
save_cursor_before_del endp
;
restore_cursor_after_del proc
        push_m  <ax,bx,cx,dx>
        mov     ah,2                    ; set crsr pos
        mov     bh,0                    ; page 0
        mov     dx,cs:crsr_pos_del      ; get saved position
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
restore_cursor_after_del endp
;
save_crsr_word  proc                    ; save cursor before ^F and ^B cmnds
        push_m  <dx,cx,bx,ax>
        mov     ah,3                    ; get crsr pos
        mov     bh,0                    ; page 0
        video_int
        mov     word ptr cs:crsr_pos_word,dx ; save position
        pop_m   <ax,bx,cx,dx>
        ret 
save_crsr_word endp
;
restore_crsr_word proc
        push_m  <ax,bx,cx,dx>
        mov     ah,2                    ; set crsr pos
        mov     bh,0                    ; page 0
        mov     dx,cs:crsr_pos_word     ; get saved position
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
restore_crsr_word endp
;
get_key proc

                                        ; get scan code and ascii code into 
                                        ; ah,al, without harming BIOS buffer.
freddy:

        cli	                                   
        mov     bx,word ptr ds:[001ah]  ; Wait for key if not there yet
        cmp     bx,word ptr ds:[001ch]                          
        jnz     entry_in
        sti
        jmp     freddy

entry_in:

        mov     ax,[bx]                 ; get scan code and ascii code                          
        inc     bx                                 
        inc     bx                                 
        mov     word ptr ds:[1ah],bx    ; inc buffer pointers                         
        cmp     bx,word ptr ds:[0082h]  ; end of buffer                        
        jnz     got_key                 ; ret if not                         
        mov     bx,word ptr ds:[0080h]  ; else loop round again in buffer
        mov     word ptr ds:[1ah],bx                         

got_key:

        ret

get_key         endp
;
cursor_off      proc                    ; turn off cursor
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get crsr size into ch,cl
        or      ch,00100000b            ; set bit 5 of ch = crsr off
        mov     ah,1
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
cursor_off      endp
;
cursor_on       proc                    ; turn the cursor back on
        push_m  <ax,bx,cx,dx>
        mov     ah,3
        mov     bh,0
        video_int                       ; get crsr size into ch,cl
        and     ch,11011111b            ; clr bit 5 of ch = crsr on
        mov     ah,1
        video_int
        pop_m   <dx,cx,bx,ax>
        ret
cursor_on       endp
;
toggle_insert_state   proc
        cmp     byte ptr cs:insert_state,0 ; is it off
        je      turn_insert_on             ; yes, so go and turn it on
        mov     byte ptr cs:insert_state,0 ; it was on, so turn off

        push    cx
        mov     cx,word ptr cs:crsr_size
        mov     ah,1
        video_int                       ; tidy up cursor size
        pop     cx
        mov     byte ptr cs:crsr_size_changed,0

        jmp     short insert_toggled

turn_insert_on:

        mov     byte ptr cs:insert_state,1
        push_m  <ax,bx,cx,dx>           ; increase crsr size
        mov     ah,3
        mov     bh,0
        video_int                       ; crsr start/end into ch,cl
        mov     word ptr cs:crsr_size,cx; and save it
        mov     byte ptr cs:crsr_size_changed,1 
        dec     ch
        dec     ch                      ; make cursor 2 scan lines higher
        mov     ah,1
        video_int                       ; set new crsr size
        pop_m   <dx,cx,bx,ax>

insert_toggled:

        ret

toggle_insert_state   endp

;
; ------------------------------------------------------------------------
;                             Data Area
; ------------------------------------------------------------------------

prompt_flag       db  0
after_prompt      db  0    ; pos of 1st char after crsr.  used by tab_key proc
ctrl_w_save       dw  0    ; saves cursor position during ctrl-w proc
unable            db  0    ; set by crsr movement routines to signify error
we_are_res        db  0    ; set to 5 by Int 16h function 2
cursor_pos        dw  0    ; save cursor position before entering editor
crsr_pos_del      dw  0    ; save cursor position while processing DEL char
crsr_pos_word     dw  0    ; save crsr pos while doing ^F, ^B and ^W cmnds
crsr_size         dw  0    ; save cursor size before changing for ins mode
crsr_size_changed db  0    ; = 1 if crsr size needs restoring
flag              db  0    ; used in cursor_left and insert procs
insert_state      db  0    ; 0 = insert mode off.  1 = on. 
pointer           db  0    ; next char to be stuffed. Points inside buffer

                           ; Scan code table for ascii characters 0 to 127
scans db 0,0,0,0,0,0,0,0,0,0,0,0,0,28,0,0,0
      db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0                      ; 0-31
      db 57,2,3,4,5,6,8,40,10,11,9,13,51,12,52,53
      db 11,2,3,4,5,6,7,8,9,10,39,39,51,13,52,53,40
      db 30,48,46,32,18,33,34,35,23,36,37,38,50,49,24,25,16 ; A-Q
      db 19,31,20,22,47,17,45,21,44                         ; R-Z
      db 26,43,27,7,12,40
      db 30,48,46,32,18,33,34,35,23,36,37,38,50,49,24,25,16 ; a-q
      db 19,31,20,22,47,17,45,21,44                         ; r-z
      db 26,43,27,40,0

; ------------------------------------------------------------------------
;                   Nothing Below Here Gets Made Resident
; ------------------------------------------------------------------------


                                        ; Note that the install routine does
                                        ; NOT shift the resident code down to
                                        ; free the space used by the program's
                                        ; PSP.  This is because part of the 
                                        ; PSP (from offset 80h) is used by
                                        ; the Ctrl-W routine as a buffer.

init:   
        jmp     install

    wrong_dos   db 'Incorrect DOS - version 2.0 or above required.',CR,LF,EOM
  save_ica_byte db 0
     slash_n    db 0

msg_header db 'ͻ',CR,LF
           db ' MS-DOS Full Screen Editor  Version '     ; no crlf here
           db VER1,'.',VER2,VER3,' (c)RJS ',CR,LF
           db 'Ķ',CR,LF,EOM
msg_ready  db '         FSE is now installed and active        ',CR,LF,EOM
msg_not_again db '       FSE seems to be already installed        ',CR,LF,EOM
; 
msg_keys1 db '             Ŀ              ',CR,LF
          db 'ĴEditor Control KeysĶ',CR,LF
          db '                           ',CR,LF
          db ' ',UP,'           Enter editor                       ',CR,LF
          db ' ',EOM

msg_keys2 db UP,DOWN,RIGHT,'        Move cursor                        ',CR,LF
           db ' RETURN      Grab line from cursor onwards      ',CR,LF
           db ' ESC         Abandon editor and restore cursor  ',CR,LF
           db ' Del         Delete current character           ',CR,LF
           db ' Backspace   Delete character to left of cursor ',CR,LF
           db ' Ins         Toggle insert mode                 ',CR,LF
           db ' Ctrl-E      Erase to end of line               ',CR,LF
           db ' Ctrl-F      Move cursor forward to next word   ',CR,LF
           db ' Ctrl-B      Move backward to previous word     ',CR,LF
           db ' Ctrl-W      Delete current word and close gap  ',CR,LF
           db ' Home        Move to left of current line       ',CR,LF
           db ' End         Move to end of current line        ',CR,LF
           db ' PgUp/PgDn   Cursor to top/bottom of screen     ',CR,LF
           db ' TAB         Cursor to character after prompt   ',CR,LF,EOM

msg_bottom db 'ͼ',CR,LF,EOM

install:

                                        ; Put FSE into memory.  Note that this
                                        ; routine does NOT shift the resident
                                        ; code down to overwrite the old PSP.
                                        ; This is because the PSP space from
                                        ; offset 80h is used as a buffer by
                                        ; various routines.


        push    cs                      ; set ds = cs
        pop     ds


                                        ; if user typed /F then go straight to
                                        ; installation, else check first to
                                        ; see if FSE is already installed.

        push    cx
        mov     cl,ds:80h               ; get length of command line params
        cmp     cl,3                    ; must be 3, for space,/,F
        pop     cx
        jne     no_flag

        push    ax
        mov     al,ds:82h               ; 1st char after space should be '/'
        cmp     al,'/'
        pop     ax
        jne     no_flag

        push    ax
        mov     al,ds:83h               ; should be f or F
        or      al,00100000b            ; force it to lower case

        cmp     al,'n'
        jne     no_n
        mov     byte ptr cs:slash_n,1   ; set flag
        pop     ax
        jmp     short not_installed2    ; jump straight to install routine,
                                        ; as using /N does an implied /F too.
no_n:

        cmp     al,'f'
        pop     ax
        jne     no_flag                 ; no /F typed
        jmp     short not_installed2

no_flag:

                                        ; If FSE is resident, calling Int 16h
                                        ; function 2 will have the side-effect
                                        ; of incrementing the 1st byte of the
                                        ; ICA, at 0000:04F0h. Test this to see
                                        ; if FSE is installed.


        push    ds
        push    ax
        mov     ax,0
        mov     ds,ax
        mov     al,ds:[04f0h]           ; get ICA byte
        mov     cs:save_ica_byte,al     ; preserve byte

        mov     ah,2                    ; if FSE is installed, this operation
        int     16h                     ; will increment 0:4f0h

        mov     ah,byte ptr ds:[04f0h]  ; get perhaps-incremented byte
        dec     ah
        cmp     ah,cs:save_ica_byte     ; was it incremented?

        jne     not_installed           ; no, so FSE is not yet installed.

        dec     byte ptr ds:4f0h        ; put byte back to original value
        pop     ax
        pop     ds

        cmp     byte ptr cs:slash_n,1
        je      n1

        show    msg_header
        show    msg_not_again
        show    msg_keys1
        disp_left
        show    msg_keys2
        show    msg_bottom
n1:
        int     20h                     ; terminate - don't reinstall


not_installed:

        pop     ax
        pop     ds

not_installed2:

        push    ax
        mov     ah,30h
        int     21h                     ; get DOS version number
        cmp     al,2
        pop     ax
        jae     dosver_ok               ; continue if at least 2

        push    ax
        push    dx
        mov     ah,9
        mov     dx,offset cs:wrong_dos  ; display message
        int     21h
        pop     dx
        pop     ax
        int     20h

dosver_ok:

        cmp     byte ptr cs:slash_n,1   ; skip the intro if user asked to
        je      n2

        show    msg_header
        show    msg_ready
        show    msg_keys1
        disp_left                       ; can't use DOS to display this char
        show    msg_keys2
        show    msg_bottom

n2:

        push    bx
        push    es
        push    ax
        mov     ah,35h                  ; get vector
        mov     al,16h                  ; for int 16h
        int     21h
        mov     word ptr cs:old_off,bx  ; pointer to old code
        mov     word ptr cs:old_seg,es
        pop     ax
        pop     es
        pop     bx

        mov     ah,25h                  ; set vector
        mov     al,16h                  ; set vector 16
        mov     dx,offset cs:intent
        int     21h

        push    ax
        push    es
        push    bx

        mov     ah,35h 
        mov     al,21h                  ; get int 21h vector
        int     21h
        mov     ax,es
        mov     cs:int21_cs,ax
        mov     cs:int21_ip,bx
        mov     ah,25h
        mov     al,21h                  ; set int 21h vector
        push    cs
        pop     ds
        mov     dx,offset cs:int21_entry
        int     21h

        mov     byte ptr cs:PSP_OFF,0   ; put null at start of buffer

        pop     bx
        pop     es
        pop     ax
    
        mov     ax,3100h                ; terminate and stay resident
        mov     dx,(init-start)/16 +17  ; size of residency in paragraphs
        int     21h                     ; done

        code    ends
        end     start                   ; start execution at label 'start'

; ------------------------------------------------------------------------
;                        (c)Robert Schifreen 1987
;
;                                The End
; ------------------------------------------------------------------------

